// @autor: jeremie@wizzilab.com
// @date: 2017-05-02

#include "DebouncedInterrupt.h"
#include "modem_ref_helper.h"
#include "modem_callbacks.h"
#include "files.h"


Semaphore button_user(0);
Semaphore modem_ready(0);
Semaphore user_ready(0);
Queue<void, 8> modem_resp;

uint8_t id, g_report_id;

enum {
    MODEM_RESP_NO,
    MODEM_RESP_TERMINAL,
    MODEM_RESP_ERROR,
    MODEM_RESP_ACK,
    MODEM_RESP_TIMEOUT,
};

alp_d7a_itf_t alarm_itf = {
    .type                           = ALP_ITF_TYPE_D7A,
    .cfg.to                         = 0,
    .cfg.te                         = 0,
    .cfg.qos.bf.resp                = D7A_RESP_ALL,
    .cfg.qos.bf.retry               = ALP_RPOL_ONESHOT,
    .cfg.addressee.ctrl.bf.nls      = D7A_NLS_AES_CCM_64,
    .cfg.addressee.ctrl.bf.idf      = D7A_ID_NBID,
    .cfg.addressee.xcl.bf           = {.s = 2, .m = 0x1},// XXX D7A_XCL_GW,
    .cfg.addressee.id[0]            = D7A_CTF_ENCODE(8),
};

// Interrupt Service Routine on button press.
void button_push_isr( void )
{
    button_user.release();
}

void print_status(uint8_t id, int status)
{
    switch (status)
    {
        case ALP_ERR_NONE:
            PRINT("Status[%d]: OK\n", id);
            break;
        default:
            PRINT("Status[%d]: error %d\n", id, status);
            break;
    }
}

// Main Callback
void my_main_callback(uint8_t terminal, int8_t err, uint8_t id)
{
    if (ALP_ERR_NONE != err)
    {
        modem_print_error(ALP_ITF_TYPE_D7A, err);
    }
    
    if (terminal)
    {
        modem_ready.release();
    }
}

// Response Callback
void my_response_callback(uint8_t terminal, int8_t err, uint8_t id)
{
    (void)id;
    
    if (ALP_ERR_NONE != err)
    {
        modem_print_error(ALP_ITF_TYPE_D7A, err);
    }
    
    if (terminal)
    {
        modem_resp.put((void*)MODEM_RESP_TERMINAL);
    }
    else
    {
        if (ALP_ERR_NONE == err)
        {
            modem_resp.put((void*)MODEM_RESP_ACK);
        }
        else
        {
            modem_resp.put((void*)MODEM_RESP_ERROR);
        }
    }
    
    // Wait end of user processing before resuming
    user_ready.wait();
}

void button_user_thread()
{
    FPRINT("(id:0x%08x)\r\n", osThreadGetId());
    
    osEvent evt;
    uint32_t resp;
    uint8_t alarm;
    uint8_t id = modem_get_id(my_main_callback);
    d7a_sp_res_t istat;
    uint8_t nb = 0;
    
    memset(&istat, 0, sizeof(d7a_sp_res_t));
    
    // Load alarm value
    ram_fs_read(FID_ALARM, 0, 1, &alarm);
    
    // Send initial value
    modem_send_file_content((uint8_t*)&alarm_itf, D7_ITF_SIZE(&alarm_itf), (void*)&istat, FID_ALARM, &alarm, 0, 1, id);
    modem_ready.wait();
    
    // Change callback
    modem_free_id(id);
    id = modem_get_id(my_response_callback);
    
    while (true)
    {
        // Wait for button press
        PRINT("PRESS BUTTON TO SEND ALARM...\r\n");
        button_user.wait();
        
        nb = 0;
        
        // load/save value to keep choerency in case of remote access...
        ram_fs_read(FID_ALARM, 0, 1, &alarm);

        // Toggle alarm state
        alarm = !alarm;
        
        ram_fs_write(FID_ALARM, 0, 1, &alarm);
        
        PRINT("BUTTON ALARM %d\r\n", alarm);
        
        // Send data
        modem_send_file_content((uint8_t*)&alarm_itf, D7_ITF_SIZE(&alarm_itf), (void*)&istat, FID_ALARM, &alarm, 0, 1, id);
        
        do
        {
            // Wait for callback
            evt = modem_resp.get(3000);
            resp = (evt.status == osEventMessage)? (uint32_t)evt.value.p : MODEM_RESP_TIMEOUT;
                        
            if (MODEM_RESP_ACK == resp)
            {
                // Print metadata
                PRINT("ACK %d: ", ++nb);
                PRINT_DATA("UID:", "%02X", istat.addressee.id, 8, " ");
                PRINT("SNR: %ddB RXLEV: %ddBm LB: %ddB\n", istat.snr, -istat.rxlev, istat.lb);
                
                // Clear istatus buffer
                memset(&istat, 0, sizeof(d7a_sp_res_t));
                
                // Resume processing
                user_ready.release();
            }
            else if (MODEM_RESP_TIMEOUT == resp)
            {
                // Could be because of the RF duty cycle,
                // especialy after boot where the duty credit is near 0.
                // The only thing to do is wait.
                // The transmission will resume once some duty credit is available.
                PRINT("WAITING...\n");
            }
            else if (MODEM_RESP_ERROR == resp)
            {
                PRINT("ERROR.\n");
                
                // Resume processing
                user_ready.release();
            }
            else if (MODEM_RESP_TERMINAL == resp)
            {
                PRINT("DONE.\n");
                
                // Resume processing
                user_ready.release();
            }

        } while (MODEM_RESP_TERMINAL != resp);
        
        PRINT("BUTTON ALARM DONE\r\n");
    }
}

// Misc
// ============================================================{{{

void my_get_alp_file_props(uint8_t fid, alp_file_header_t* hdr)
{
    memcpy(hdr, ram_fs_get_header(fid), sizeof(alp_file_header_t));
}

modem_callbacks_t callbacks = {
    .read       = my_read,
    .write      = my_write,
    .read_fprop = my_read_fprop,
    .flush      = my_flush,
    .remove     = my_delete,
    .udata      = my_udata,
    .lqual      = my_lqual,
    .ldown      = my_ldown,
    .reset      = my_reset,
    .boot       = my_boot,
    .busy       = my_busy,
};


/*** Main function ------------------------------------------------------------- ***/
int main()
{
    // Start & initialize
#ifdef DEBUG_LED
    DBG_OPEN(DEBUG_LED);
#else
    DBG_OPEN(NC);
#endif
    PRINT("\n"
          "-----------------------------------------\n"
          "---------- Demo Send file data ----------\n"
          "-----------------------------------------\n");
              
    FPRINT("(id:0x%08x)\r\n", osThreadGetId());
    
    modem_helper_open(&callbacks);
    
    uint8_t id = modem_get_id(my_main_callback);
    
    PRINT("Register Files\n");
    // Button/Alarm is a local file. As we want to check the outcome of sending
    // this, don't use D7AActP Notification but rather plain ALP ITF Forwarding.
    // Declaration just allows remote access.
    modem_update_file(FID_ALARM, (alp_file_header_t*)&h_alarm, (uint8_t*)&f_alarm);

    // Configure URC: LQUAL on report file notification every 10 reports
    PRINT("Setup URCs\n");
    modem_enable_urc(ALP_URC_TYPE_LQUAL, IFID_REPORT, 10, true, id);
    modem_ready.wait();
    
    PRINT("Start D7A Stack\n");
    modem_activate_itf(ALP_ITF_TYPE_D7A, 24, 0, ALP_D7A_ISTAT_RESP | ALP_D7A_ISTAT_UNS | ALP_D7A_ISTAT_EOP, true, id);
    modem_ready.wait();
    
    PRINT("Notify Modem Version\n");
    modem_notify_file(D7A_FID_FIRMWARE_VERSION, 0, SIZE_HOST_REV, id);
    modem_ready.wait();
    
    PRINT("Notify FW Version\n");
    uint8_t default_root_key[16] = DEFAULT_ROOT_KEY;
    modem_notify_host_rev(&f_rev, &h_rev, default_root_key);
    
    modem_free_id(id);

#ifdef DEBUG_BUTTON
    DebouncedInterrupt user_interrupt(DEBUG_BUTTON);
    user_interrupt.attach(button_push_isr, IRQ_FALL, 500, true);
    
    Thread but_th(osPriorityNormal, 2048, NULL);
    osStatus status = but_th.start(button_user_thread);
    ASSERT(status == osOK, "Failed to start but thread (err: %d)\r\n", status);
#endif

#ifdef DEBUG_LED
    DigitalOut my_led(DEBUG_LED);
#endif
    
    // Set main task to lowest priority
    osThreadSetPriority(osThreadGetId(), osPriorityIdle);
    while(true)
    {
        Thread::wait(500);
#ifdef DEBUG_LED
        my_led = !my_led;
#endif
    }
}